/*
   Copyright (C) 2004 by James Gregory
   Part of the GalaxyHack project
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License.
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.
 
   See the COPYING file for more details.
*/

#include "JSDL.h"
#include "Logging.h"
#include "Globals.h"

#include <boost/filesystem/operations.hpp>
#include <SDL_image.h>
#include <SDL_rotozoom.h>
#include <SDL_mixer.h>

#include <cstdlib>
#include <fstream>
#include <stdexcept>

using std::runtime_error;
using std::memcpy;

//defines from SDL_gfx
#define clip_xmin(surface) surface->clip_rect.x
#define clip_xmax(surface) surface->clip_rect.x+surface->clip_rect.w-1
#define clip_ymin(surface) surface->clip_rect.y
#define clip_ymax(surface) surface->clip_rect.y+surface->clip_rect.h-1

void JamesSDL::Init() {	
	screenRect.x = 0;
	screenRect.y = 0;
	screenRect.w = globalSettings.screenWidth;
	screenRect.h = globalSettings.screenHeight;
	
	if (!globalSettings.batch) {
		if (SDL_Init(SDL_INIT_VIDEO) == -1) {
			char output[120];
			sprintf(output, SDL_GetError());
			throw runtime_error(output);
		}
		
		if (!globalSettings.disableSound) {
			if (SDL_InitSubSystem(SDL_INIT_AUDIO) == -1) {
				char error[120];
				sprintf(error, SDL_GetError());
				string errorStr = error;
				errorStr += ", continuing without initialising sound.";
				WriteLog(errorStr);
				//set to false in SettingsStruct constructor
				globalSettings.dontWriteSound = true;
				globalSettings.disableSound = true;
			}
		}
	
		if (globalSettings.fullScreen)
			screen = SDL_SetVideoMode(globalSettings.screenWidth, globalSettings.screenHeight, screenBPP, SDL_HWSURFACE | SDL_DOUBLEBUF | SDL_FULLSCREEN);
		else
			screen = SDL_SetVideoMode(globalSettings.screenWidth, globalSettings.screenHeight, screenBPP, SDL_HWSURFACE | SDL_DOUBLEBUF);// | SDL_NOFRAME);
	
		if ( screen == NULL ) {
			SDL_Quit();
			char output[120];
			sprintf(output, SDL_GetError());
			throw runtime_error(output);
		}
	}

	//keyboard
	SDL_EnableUNICODE(1);
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
	keyboard = SDL_GetKeyState(0);

	//no mouse cursor with DGA so we may as well always do it ourself
	SDL_ShowCursor(SDL_DISABLE);

	if (!globalSettings.disableSound) {
		//4096 is rather a large chunksize, should be smaller if I add sound effects
		if (Mix_OpenAudio(MIX_DEFAULT_FREQUENCY, MIX_DEFAULT_FORMAT, 2, 4096) == -1) {
			char output[100];
			sprintf(output, Mix_GetError());
			throw runtime_error(output);
		}
	}
}

void JamesSDL::BitmapHDtoFile(const string& filename) {
	if (globalSettings.batch)
		return;
		
	loadingSurface = IMG_Load(filename.c_str());

	if (loadingSurface == NULL) {
		char output[100];
		sprintf(output, IMG_GetError());
		throw runtime_error(output);
	}

	if (loadingSurface->format->BitsPerPixel != screenBPP)
		loadingSurface = SDL_ConvertSurface(loadingSurface, screen->format, SDL_SWSURFACE);
}

void JamesSDL::BitmapColorConvert(Uint16 newColor) {
	if (globalSettings.batch)
		return;
		
	Uint16 pink = MapRGB(255, 0, 255);

	SDL_LockSurface(loadingSurface);

	Uint16* castedBuffer = reinterpret_cast<Uint16*>(loadingSurface->pixels);

	for (int i = 0; i != (loadingSurface->pitch / 2) * loadingSurface->h; ++i) {
		if (castedBuffer[i] == pink)
			castedBuffer[i] = newColor;
	}

	SDL_UnlockSurface(loadingSurface);
}

void JamesSDL::FlipHorizontal() {
	if (globalSettings.batch)
		return;
		
	int nBytes = loadingSurface->pitch * loadingSurface->h * 2;
	char* buffer = new char[nBytes];

	SDL_LockSurface(loadingSurface);

	//copy image to work area
	//we rely here on the fact that memCpy increments the pointer it uses for copying
	//using ++ or some such rather than +ing it by 1 byte, even though it expects char
	//pointers
	//this means that we must give it the number of pixels rather than the number of bytes
	memcpy(buffer, loadingSurface->pixels, nBytes / 2);

	Uint16* pSource = reinterpret_cast<Uint16*>(buffer);
	Uint16* pDest = reinterpret_cast<Uint16*>(loadingSurface->pixels);

	//slightly strange indices
	//flip horizontally
	for (int i = 1; i != loadingSurface->h + 1; ++i) {
		for (int j = 1; j != loadingSurface->w + 1; ++j) {
			*pDest = pSource[i * loadingSurface->w - j];
			++pDest;
		}
	}

	SDL_UnlockSurface(loadingSurface);

	// release the memory
	delete[] buffer;
}

void JamesSDL::BitmapCloseFile() {
	if (globalSettings.batch)
		return;
		
	SDL_FreeSurface(loadingSurface);
	loadingSurface = 0;
}

SDL_Surface* JamesSDL::BitmapFiletoSurface(int x, int y, int width, int height, Uint16 transparency) {
	SDL_Surface* ret = SDL_CreateRGBSurface(SDL_HWSURFACE, width, height, screenBPP,  0, 0, 0, 0);
	
	if (globalSettings.batch)
		return ret;

	SDL_Rect tempRect = {x, y, width, height};

	if (width - x != loadingSurface->w || height - y != loadingSurface->h) {
		double zoomx = static_cast<double>(width - x) / loadingSurface->w;
		double zoomy = static_cast<double>(height - y) / loadingSurface->h;

		//FIXME FIXME FIXME utter hackery of the highest order, without this
		//medium capital ships end up being 511*111
		zoomx+= 0.00000000005;
		zoomy+= 0.00000000005;

		SDL_Surface* tempSurface = zoomSurface (loadingSurface, zoomx, zoomy, 1);

		SDL_BlitSurface(tempSurface, &tempRect, ret, 0);

		SDL_FreeSurface(tempSurface);
	} else
		//Just blit across
		SDL_BlitSurface(loadingSurface, &tempRect, ret, 0);


	SDL_SetColorKey(ret, SDL_SRCCOLORKEY | SDL_RLEACCEL, transparency);

	return ret;
}

SDL_Surface* JamesSDL::BitmapHDtoSurface(const string& filename, int x, int y, int w, int h, Uint16 transparency) {
	BitmapHDtoFile(filename);
	SDL_Surface* ret = BitmapFiletoSurface(x, y, w, h, transparency);
	BitmapCloseFile();

	return ret;
}

//////

void JamesSDL::BltFill(SDL_Rect destRect, Uint16 fillColor) {
	if (skipDisplayFrame || globalSettings.batch)
		return;		
	if (SDL_FillRect(screen, &destRect, fillColor) == -1)
		throw runtime_error("Failed to BltFill");
}

void JamesSDL::Blt(SDL_Surface* src, SDL_Rect destRect) {
	if (skipDisplayFrame || globalSettings.batch)
		return;	
	if (SDL_BlitSurface(src, 0, screen, &destRect) == -1)
		throw runtime_error("Failed to blit");
}


void JamesSDL::BltPart(SDL_Surface* src, SDL_Rect& srcRect, SDL_Rect destRect) {
	if (skipDisplayFrame || globalSettings.batch)
		return;	
	if (SDL_BlitSurface(src, &srcRect, screen, &destRect) == -1)
		throw runtime_error("BltPart fails");
}

void JamesSDL::Flip() {
	if (skipDisplayFrame || globalSettings.batch)
		return;	
	SDL_Flip(screen);
}

void JamesSDL::LockBack() {
	if (globalSettings.batch)
		return;	
	if (SDL_LockSurface(screen) == -1)
		throw runtime_error("Failed to lock back surface");
}

void JamesSDL::UnlockBack() {
	if (globalSettings.batch)
		return;	
	SDL_UnlockSurface(screen);
}

void JamesSDL::Shutdown() {
	if (theMusic)
		Mix_FreeMusic(theMusic);

	Mix_CloseAudio();

	//this automatically frees the screen
	SDL_Quit();
}

Uint16 JamesSDL::MapRGB(int r, int g, int b) {
	return SDL_MapRGB(screen->format, r, g, b);
}

void JamesSDL::UshortToRGB(Uint16 theColor, Uint8& giveR, Uint8& giveG, Uint8& giveB) {
	SDL_GetRGB(theColor, screen->format, &giveR, &giveG, &giveB);
}

void JamesSDL::LoadMusic(const string& filename) {
	if (globalSettings.disableSound)
		return;

	musicName = filename;

	if (theMusic)
		Mix_FreeMusic(theMusic);

	theMusic = Mix_LoadMUS(filename.c_str());

	if (!theMusic) {
		char output[100];
		sprintf(output, Mix_GetError());
		throw runtime_error(output);
	}
}

void JamesSDL::PlayMusic() {
	if (!globalSettings.bMusic || globalSettings.disableSound)
		return;

	if (Mix_PlayMusic(theMusic, -1) == -1) {
		char output[100];
		sprintf(output, Mix_GetError());
		throw runtime_error(output);
	}
}

void JamesSDL::StopMusic() {
	Mix_HaltMusic();
}

/* ----- Circle */

/* Note: Based on algorithm from sge library, modified by A. Schiffler and James Gregory*/
/* with multiple pixel-draw removal and other minor speedup changes.   */

int JamesSDL::CircleColor(SDL_Surface * dst, Sint16 x, Sint16 y, Sint16 r, Uint32 color) {
    Sint16 left, right, top, bottom;
    int result;
    Sint16 x1, y1, x2, y2;
    Sint16 cx = 0;
    Sint16 cy = r;
    Sint16 ocx = (Sint16) 0xffff;
    Sint16 ocy = (Sint16) 0xffff;
    Sint16 df = 1 - r;
    Sint16 d_e = 3;
    Sint16 d_se = -2 * r + 5;
    Sint16 xpcx, xmcx, xpcy, xmcy;
    Sint16 ypcy, ymcy, ypcx, ymcx;
    Uint8 *colorptr;

    /*
     * Sanity check radius 
     */
    if (r < 0) {
	return (-1);
    }

    /*
     * Special case for r=0 - draw a point 
     */
    if (r == 0) {
	return (pixelColor(dst, x, y, color));
    }

    /*
     * Get clipping boundary 
     */
    left = dst->clip_rect.x;
    right = dst->clip_rect.x + dst->clip_rect.w - 1;
    top = dst->clip_rect.y;
    bottom = dst->clip_rect.y + dst->clip_rect.h - 1;

    /*
     * Test if bounding box of circle is visible 
     */
    x1 = x - r;
    x2 = x + r;
    y1 = y - r;
    y2 = y + r;
    if ((x1<left) && (x2<left)) {
     return(0);
    } 
    if ((x1>right) && (x2>right)) {
     return(0);
    } 
    if ((y1<top) && (y2<top)) {
     return(0);
    } 
    if ((y1>bottom) && (y2>bottom)) {
     return(0);
    } 

    /*
     * Draw circle 
     */
    result = 0;

	/*
	 * No Alpha - direct memory writes 
	 */

	/*
	 * Setup color 
	 */
	colorptr = (Uint8 *) & color;
	if (SDL_BYTEORDER == SDL_BIG_ENDIAN) {
	    color = SDL_MapRGBA(dst->format, colorptr[0], colorptr[1], colorptr[2], colorptr[3]);
	} else {
	    color = SDL_MapRGBA(dst->format, colorptr[3], colorptr[2], colorptr[1], colorptr[0]);
	}

	/*
	 * Draw 
	 */
	do {
	    if ((ocy != cy) || (ocx != cx)) {
		xpcx = x + cx;
		xmcx = x - cx;
		if (cy > 0) {
		    ypcy = y + cy;
		    ymcy = y - cy;
		    result |= fastPixelColorNolock(dst, xmcx, ypcy, color);
		    result |= fastPixelColorNolock(dst, xpcx, ypcy, color);
		    result |= fastPixelColorNolock(dst, xmcx, ymcy, color);
		    result |= fastPixelColorNolock(dst, xpcx, ymcy, color);
		} else {
		    result |= fastPixelColorNolock(dst, xmcx, y, color);
		    result |= fastPixelColorNolock(dst, xpcx, y, color);
		}
		ocy = cy;
		xpcy = x + cy;
		xmcy = x - cy;
		if (cx > 0) {
		    ypcx = y + cx;
		    ymcx = y - cx;
		    result |= fastPixelColorNolock(dst, xmcy, ypcx, color);
		    result |= fastPixelColorNolock(dst, xpcy, ypcx, color);
		    result |= fastPixelColorNolock(dst, xmcy, ymcx, color);
		    result |= fastPixelColorNolock(dst, xpcy, ymcx, color);
		} else {
		    result |= fastPixelColorNolock(dst, xmcy, y, color);
		    result |= fastPixelColorNolock(dst, xpcy, y, color);
		}
		ocx = cx;
	    }
	    /*
	     * Update 
	     */
	    if (df < 0) {
		df += d_e;
		d_e += 2;
		d_se += 2;
	    } else {
		df += d_se;
		d_e += 2;
		d_se += 4;
		cy--;
	    }
	    cx++;
	} while (cx <= cy);


    return (result);
}

int JamesSDL::fastPixelColorNolock(SDL_Surface * dst, Sint16 x, Sint16 y, Uint32 color) {
    int bpp;
    Uint8 *p;

    /*
     * Honor clipping setup at pixel level 
     */
    if ((x >= clip_xmin(dst)) && (x <= clip_xmax(dst)) && (y >= clip_ymin(dst)) && (y <= clip_ymax(dst))) {

	/*
	 * Get destination format 
	 */
	bpp = dst->format->BytesPerPixel;
	p = (Uint8 *) dst->pixels + y * dst->pitch + x * bpp;
	switch (bpp) {
	case 1:
	    *p = color;
	    break;
	case 2:
	    *(Uint16 *) p = color;
	    break;
	case 3:
	    if (SDL_BYTEORDER == SDL_BIG_ENDIAN) {
		p[0] = (color >> 16) & 0xff;
		p[1] = (color >> 8) & 0xff;
		p[2] = color & 0xff;
	    } else {
		p[0] = color & 0xff;
		p[1] = (color >> 8) & 0xff;
		p[2] = (color >> 16) & 0xff;
	    }
	    break;
	case 4:
	    *(Uint32 *) p = color;
	    break;
	}			/* switch */


    }

    return (0);
}

int JamesSDL::pixelColor(SDL_Surface * dst, Sint16 x, Sint16 y, Uint32 color)
{
    Uint8 alpha;
    Uint32 mcolor;
    int result = 0;

    /*
     * Setup color 
     */
    alpha = color & 0x000000ff;
    mcolor =
	SDL_MapRGBA(dst->format, (color & 0xff000000) >> 24,
		    (color & 0x00ff0000) >> 16, (color & 0x0000ff00) >> 8, alpha);

    /*
     * Draw 
     */
    result = _putPixelAlpha(dst, x, y, mcolor, alpha);

    return (result);
}

/* New, faster routine - default blending pixel */

int JamesSDL::_putPixelAlpha(SDL_Surface * surface, Sint16 x, Sint16 y, Uint32 color, Uint8 alpha) {
    Uint32 Rmask = surface->format->Rmask, Gmask =
	surface->format->Gmask, Bmask = surface->format->Bmask, Amask = surface->format->Amask;
    Uint32 R, G, B, A = 0;

    if (x >= clip_xmin(surface) && x <= clip_xmax(surface)
	&& y >= clip_ymin(surface) && y <= clip_ymax(surface)) {

	switch (surface->format->BytesPerPixel) {
	case 1:{		/* Assuming 8-bpp */
		if (alpha == 255) {
		    *((Uint8 *) surface->pixels + y * surface->pitch + x) = color;
		} else {
		    Uint8 *pixel = (Uint8 *) surface->pixels + y * surface->pitch + x;

		    Uint8 dR = surface->format->palette->colors[*pixel].r;
		    Uint8 dG = surface->format->palette->colors[*pixel].g;
		    Uint8 dB = surface->format->palette->colors[*pixel].b;
		    Uint8 sR = surface->format->palette->colors[color].r;
		    Uint8 sG = surface->format->palette->colors[color].g;
		    Uint8 sB = surface->format->palette->colors[color].b;

		    dR = dR + ((sR - dR) * alpha >> 8);
		    dG = dG + ((sG - dG) * alpha >> 8);
		    dB = dB + ((sB - dB) * alpha >> 8);

		    *pixel = SDL_MapRGB(surface->format, dR, dG, dB);
		}
	    }
	    break;

	case 2:{		/* Probably 15-bpp or 16-bpp */
		if (alpha == 255) {
		    *((Uint16 *) surface->pixels + y * surface->pitch / 2 + x) = color;
		} else {
		    Uint16 *pixel = (Uint16 *) surface->pixels + y * surface->pitch / 2 + x;
		    Uint32 dc = *pixel;

		    R = ((dc & Rmask) + (((color & Rmask) - (dc & Rmask)) * alpha >> 8)) & Rmask;
		    G = ((dc & Gmask) + (((color & Gmask) - (dc & Gmask)) * alpha >> 8)) & Gmask;
		    B = ((dc & Bmask) + (((color & Bmask) - (dc & Bmask)) * alpha >> 8)) & Bmask;
		    if (Amask)
			A = ((dc & Amask) + (((color & Amask) - (dc & Amask)) * alpha >> 8)) & Amask;

		    *pixel = R | G | B | A;
		}
	    }
	    break;

	case 3:{		/* Slow 24-bpp mode, usually not used */
		Uint8 *pix = (Uint8 *) surface->pixels + y * surface->pitch + x * 3;
		Uint8 rshift8 = surface->format->Rshift / 8;
		Uint8 gshift8 = surface->format->Gshift / 8;
		Uint8 bshift8 = surface->format->Bshift / 8;
		Uint8 ashift8 = surface->format->Ashift / 8;


		if (alpha == 255) {
		    *(pix + rshift8) = color >> surface->format->Rshift;
		    *(pix + gshift8) = color >> surface->format->Gshift;
		    *(pix + bshift8) = color >> surface->format->Bshift;
		    *(pix + ashift8) = color >> surface->format->Ashift;
		} else {
		    Uint8 dR, dG, dB, dA = 0;
		    Uint8 sR, sG, sB, sA = 0;

		    pix = (Uint8 *) surface->pixels + y * surface->pitch + x * 3;

		    dR = *((pix) + rshift8);
		    dG = *((pix) + gshift8);
		    dB = *((pix) + bshift8);
		    dA = *((pix) + ashift8);

		    sR = (color >> surface->format->Rshift) & 0xff;
		    sG = (color >> surface->format->Gshift) & 0xff;
		    sB = (color >> surface->format->Bshift) & 0xff;
		    sA = (color >> surface->format->Ashift) & 0xff;

		    dR = dR + ((sR - dR) * alpha >> 8);
		    dG = dG + ((sG - dG) * alpha >> 8);
		    dB = dB + ((sB - dB) * alpha >> 8);
		    dA = dA + ((sA - dA) * alpha >> 8);

		    *((pix) + rshift8) = dR;
		    *((pix) + gshift8) = dG;
		    *((pix) + bshift8) = dB;
		    *((pix) + ashift8) = dA;
		}
	    }
	    break;

	case 4:{		/* Probably 32-bpp */
		if (alpha == 255) {
		    *((Uint32 *) surface->pixels + y * surface->pitch / 4 + x) = color;
		} else {
		    Uint32 *pixel = (Uint32 *) surface->pixels + y * surface->pitch / 4 + x;
		    Uint32 dc = *pixel;

		    R = ((dc & Rmask) + (((color & Rmask) - (dc & Rmask)) * alpha >> 8)) & Rmask;
		    G = ((dc & Gmask) + (((color & Gmask) - (dc & Gmask)) * alpha >> 8)) & Gmask;
		    B = ((dc & Bmask) + (((color & Bmask) - (dc & Bmask)) * alpha >> 8)) & Bmask;
		    if (Amask)
			A = ((dc & Amask) + (((color & Amask) - (dc & Amask)) * alpha >> 8)) & Amask;

		    *pixel = R | G | B | A;
		}
	    }
	    break;
	}
    }

    return (0);
}
